import re
import threading

import discord
import discord.ext.commands as commands
from discord.ext import tasks

from src.steam_service import get_server_list, get_populated_servers


class Notification(commands.Cog):
    def __init__(self, bot, config):
        self.bot = bot
        self.config = config

        self.should_ping_role = True
        self.lock_emojis = {
            # Keys are the boolean int values returned by "password_protected"
            1: discord.PartialEmoji(name=":lock:"),
            0: discord.PartialEmoji(name=":unlock:")
        }

        self.lock = threading.Lock()
        self.fetch_servers_info.start()

    @tasks.loop(seconds=300)
    async def fetch_servers_info(self, is_update_forced=False):
        with self.lock:
            servers = get_server_list()
            populated_servers = get_populated_servers(servers, self.config["min_pop"])

            if len(populated_servers) > 0:
                await self.send_notification(populated_servers)
                self.timeout_role_ping()
            elif is_update_forced:
                await self.send_empty_servers_notification()

    async def send_notification(self, populated_servers):
        channel = self.bot.get_channel(self.config["channel_id"])
        role = channel.guild.get_role(self.config["role_id"])

        new_msg_content = self.create_msg_content(populated_servers, role)

        # TODO Rewrite this
        last_message = await channel.fetch_message(channel.last_message_id)
        # If the message at the bottom of chat is bot's update then edit it with new values to reduce spamming
        if last_message.author.id == self.bot.user.id:
            # Editing messages dont trigger pings, so send a message if ping timeout passed
            if self.should_ping_role:
                self.last_update_msg = await channel.send(new_msg_content)
            else:
                await last_message.edit(content=new_msg_content)
        else:
            self.last_update_msg = await channel.send(new_msg_content)

            # Only timeout ping if its already on, otherwise it will spawn multiple timers to reset ping
            if self.should_ping_role:
                self.timeout_role_ping()

    def create_msg_content(self, populated_servers, role):
        if self.should_ping_role:
            message = f"{role.mention}\n People are playing on:\n\n"
        else:
            message = "People are playing on:\n\n"

        for server in populated_servers:
            server["server_name"] = self.__remove_links_from_name(server["server_name"])

            server_protected_emoji = self.lock_emojis[server["password_protected"]]
            message += f'{server["server_name"]} {server_protected_emoji} - {server["map"]} \n'
            message += f'Players: {server["player_count"]} / {server["max_players"]} \n\n'

        return message

    async def send_empty_servers_notification(self):
        channel = self.bot.get_channel(self.config["channel_id"])
        await channel.send("All servers are empty.")

    @fetch_servers_info.before_loop
    async def before_fetch_task(self):
        await self.bot.wait_until_ready()

    def timeout_role_ping(self):
        self.should_ping_role = False
        timer = threading.Timer(float(self.config["timeout"]), self.restart_role_ping)
        timer.start()

    def restart_role_ping(self):
        self.should_ping_role = True

    @commands.command()
    async def update(self, message: discord.Message):
        """Force active server check"""
        await self.fetch_servers_info(is_update_forced=True)

    # TODO There is libraries that do this better
    def __remove_links_from_name(self, server_name):
            regex = r'('

            # Scheme (HTTP, HTTPS, FTP and SFTP):
            regex += r'(?:(https?|s?ftp):\/\/)?'

            # www:
            regex += r'(?:www\.)?'

            regex += r'('

            # Host and domain (including ccSLD):
            regex += r'(?:(?:[A-Z0-9][A-Z0-9-]{0,61}[A-Z0-9]\.)+)'

            # TLD:
            regex += r'([A-Z]{2,6})'

            # IP Address:
            regex += r'|(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'

            regex += r')'

            # Port:
            regex += r'(?::(\d{1,5}))?'

            # Query path:
            regex += r'(?:(\/\S+)*)'

            regex += r')'

            find_urls = re.compile(regex, re.IGNORECASE)
            safe_name = re.sub(find_urls, "[LINK REMOVED]", server_name)

            return safe_name
